import java.io.*;
import javax.microedition.io.*;
import java.util.Vector;

public class FTPConnection implements Runnable {
	FTPMDlet mdlet;
	private boolean alive=false;
    private FTPData ftpd=null;
    private StreamConnection controlConnection;
	private final static String CRLF = "\r\n";

	public final static int STOR=2;
	public final static int RETR=1;
	public final static int LIST=3;
    public final static int IDLE=0;
	private int job=0;   
	private boolean terminating=false;
	private byte[] downloadeddata=null;
	//private String fname;
	private FTPListener ftpl=null;
	private int bytein=0, byteout=0;

    private String host, user, pass,dir;
    Thread th = null;
    //Thread dth = null;
    private Vector queue = new Vector();
    private Vector queuenames = new Vector();
    private Vector queuedownloadnames=new Vector();
    

	public FTPConnection(FTPMDlet mdlet) {
		this.mdlet = mdlet;
		ftpd = new FTPData(mdlet);
        th = new Thread(this);
       // dth = new Thread(ftpd);
	}
	public void setFTPListener(FTPListener ftpl) {this.ftpl = ftpl; ftpd.setFTPListener(ftpl);}
	public void removeFTPListener() {ftpl=null; ftpd.removeFTPListener();}

	public void connect(String host, String user, String pass, String dir) {
		if (!alive) {
			this.host = host;
			this.user = user;
			this.pass = pass;
			this.dir = dir;
		    mdlet.showLog();
		    if (!th.isAlive()) {th=new Thread(this); System.out.println("FTPC:New Thread");}
		    //if (!dth.isAlive()) { dth = new Thread(ftpd); System.out.println("FTPC:New Data Thread "+dth);}
		    //if (!ftpd.isAlive()) ftpd.startThread();
			th.start();
		}
	}
	public void disconnect() {
		if (alive ) {
			System.out.println("FTPC : Disconnect");
			//ftpd.closeDataPort();
			alive=false;				
		}
	}

	public void uploaddata(byte[] bb, String fname) {
		queue.addElement(bb);
    	queuenames.addElement(fname);

		/*bdata = bb;
		this.fname = fname;*/
		job = STOR;
	}

	public void downloaddata(String fname) {
		queuedownloadnames.addElement(fname);
		job = RETR;		
	}

	public byte[] getdownloaded() {
		downloadeddata= ftpd.download();
		return downloadeddata;
	}

	public void retrList() {
		job = LIST;
	}

	public void cancel() {
		ftpd.closeConnection();
	}

	public void run () {
		th.setPriority(Thread.MIN_PRIORITY);
		DataInputStream in=null;
		DataOutputStream out=null;        
		try {			
			controlConnection = (StreamConnection) Connector.open("socket://"+host+":21");
			System.out.println("FTPC : Connection opened");
			alive = true;
			in = controlConnection.openDataInputStream();
			out = controlConnection.openDataOutputStream();
			int reply;
			//mdlet.addLog3("Connecting...");
			mdlet.addLogState("Connecting");

			reply = readReply(in);
			if (reply/100 != 2) {throw new FTPException("Opening failed!");}
			//mdlet.addLog2(reply);
			//mdlet.addLog("USER...");
			writeString(out, "USER "+user);

            reply = readReply(in);
			if (reply/100 != 3) {throw new FTPException("USER command failed!\nCheck program settings!");}			
			//mdlet.addLog2(reply);
			//mdlet.addLog("PASS...");
            writeString(out, "PASS "+pass);

            reply = readReply(in);
			if (reply/100 != 2) {throw new FTPException("Password wrong!\nCheck program settings!");}
			//mdlet.addLog2(reply);

			writeString(out, "TYPE I");
			reply=readReply(in);
			//mdlet.addLog2(reply);
			if (reply != 200) throw new FTPException("Cannot set TYPE to Image: ");

			writeString(out, "CDUP ");
			reply=readReply(in);
			//mdlet.addLog2(reply);

            //mdlet.addLog("CWD...");
			writeString(out, "CWD "+dir); 
            reply = readReply(in);

            if (reply != 250) {
            	writeString(out, "MKD "+dir);
            	reply=readReply(in);
            	if (reply != 257) {
            		throw new FTPException("Cannot create folder!\nCheck program settings!\n(Server reply:"+reply+")");	
            	}
            	writeString(out, "CWD "+dir);
            	reply = readReply(in);
            }
			if (reply != 250) {throw new FTPException("Changing working directory failed\n"+
				"Check if program settings are correct");} 			
			//mdlet.addLog2(reply);
			mdlet.addLogCommand("Data port");
			try { ftpd.startThread(); System.out.println("FTPC:FTPD Works successfully!"); } 
			catch (IllegalThreadStateException e) {
				ftpd.terminateThread();
			/*	ftpd=null;
				ftpd=new FTPData(mdlet);*/
				/*dth=null;
				dth = new Thread(ftpd);
				dth.start();*/ftpd.startThread();
				System.out.println("FTPC:FTPD recreated!");
			}
			
            if (ftpl != null) ftpl.actionComplete(ftpl.CONNECT, null);
            mdlet.addLogState(true);
        } catch (FTPException fe) {
        	mdlet.addLogState(false);
        	mdlet.messageBox("Error!", fe.getMessage()+"\n(Server reply:"+lastreply+")"); 
			alive=false;
			return;			
        } catch (Exception e) {
			//mdlet.addLog("FTPC:NotStarted:"+e);
			mdlet.addLogState(false);
            mdlet.messageBox("Error!", e.toString()); 
			alive=false;
			return;
			//writeString(out, "QUIT");
		}

		while (alive) {
			int reply;
			switch(job){
			  case STOR : 
				try{
					byte[] bdata=(byte[])queue.firstElement();
					String fname=(String)queuenames.firstElement();
					mdlet.showLog();
					mdlet.addLogState("Uploading file '"+fname+"'");
					//mdlet.addLog("PORT...");
					writeString(out, "PORT "+ftpd.getPORT());
                    reply = readReply(in);
					if (reply/100 != 2) {throw new FTPException("Port not accepted!");}
			
					//mdlet.addLog2(reply);
					ftpd.setJob(ftpd.UPLOAD, ftpl.STORE, fname);
					ftpd.upload(bdata);
					writeString(out, "STOR "+fname);
					//mdlet.addLog("FTPC: uploading...");
					reply = readReply(in);
					//mdlet.addLog2(readReply(in));
					ftpd.startConnection();
		            //et.addLog2(readReply(in));
		            reply = readReply(in);           			                    
					ftpd.closeConnection();

                    /* remove from queue */
                    queue.removeElement(bdata);
                    queuenames.removeElement(fname);
                    if (queue.isEmpty()) {job = IDLE;} else {job = STOR;}
                    /*
                    if (ftpl != null) {
                    	//ftpl.fname = fname;
                    	//ftpl.actionComplete(ftpl.STORE, fname);
                    }*/
                    //mdlet.addLogState(true);
                } catch (Exception e) {
		        	System.out.println("FTPC:Error!(STOR)");
		        	mdlet.messageBox("Error!", e.toString());
		        }
                break;
			  case RETR :
				try{
					String fname=(String)queuedownloadnames.firstElement();
					mdlet.showLog();
					mdlet.addLogState("Downloading file '"+fname+"'");
                    //mdlet.addLog("PORT...");
					writeString(out, "PORT "+ftpd.getPORT());
					reply=readReply(in);
					//mdlet.addLog2(readReply(in));

					ftpd.setJob(ftpd.DOWNLOAD, ftpl.RETRIVE, fname);
					//mdlet.addLog("RETR...");
					writeString(out, "RETR "+fname);
                    //mdlet.addLog("FTPC: downloading...");
                    reply = readReply(in);
                    //mdlet.addLog2(readReply(in));                    
                    reply = readReply(in);
                    //mdlet.addLog2(readReply(in));                    
                    ftpd.startConnection();
                    ftpd.closeConnection();

                    queuedownloadnames.removeElement(fname);
                    if (queuedownloadnames.isEmpty()) { job = IDLE;} else {job = RETR;}
                } catch (Exception e) {
                	System.out.println("FTPC:Error!(RETR)");
                }
				break;
			  case LIST :
			  	try {
			  		mdlet.showLog();
			  		mdlet.addLogState("Retriving folder list");
                    //mdlet.addLog("PORT...");
					writeString(out, "PORT "+ftpd.getPORT());
                    reply = readReply(in);
					//mdlet.addLog2(readReply(in));

			  		ftpd.setJob(ftpd.DOWNLOAD, ftpl.LIST, null);
			  		writeString(out, "LIST ");			  		
			  		//mdlet.addLog2(readReply(in));			  		
			  		//mdlet.addLog2(readReply(in));
                    reply = readReply(in);
                    reply = readReply(in);
                    ftpd.startConnection();
			  		ftpd.closeConnection();			  		
			  		job = IDLE;
			  	} catch(Exception e) {
			  		System.out.println("FTPC:Error!(LIST):"+e);
			  	}
			  	break;		
			}
        } 
        mdlet.addLogState("Disconnecing");
        System.out.println("FTPC: qiuting...");
		try {
			writeString(out, "QUIT ");
			//mdlet.addLog("QUIT...");
		} catch (Exception e) {
			System.out.println("FTPC: server doesn't want to close!");
		}
		ftpd.terminateThread();
		try {
			if (in != null) in.close();
			if (out != null) out.close();
	        if (controlConnection != null) controlConnection.close();

        } catch (Exception e) {
			mdlet.addLogState("FTPC Error on closing: "+e);
		}
        if (ftpl != null) ftpl.actionComplete(ftpl.DISCONNECT, null);
	}	

	public int getBytesIn(){ return bytein+ftpd.bytein;	}
	public int getBytesOut() {return byteout+ftpd.byteout; }

	private String lastreply;

	private int readReply(DataInputStream in) throws IOException {		
        String data = "";
        /*do {
        	byte[] b = new byte[1000];
            int readed = in.read(b);    
            String ts = new String(b, 0, readed);
            data = data.concat(ts);
            System.out.println("Readed :" +ts);
        } while (!data.endsWith(CRLF));*/
        int repl=-1;
        do{
        	int b;
        	b = in.read();
        	data = data+(char)b;
        	bytein++;
        	if (repl == -1) try {
        		lastreply = data;
        		repl = Integer.parseInt(data.substring(0,3));
        	} catch (Exception e) {  }
        	if ((data.length() > 3) && (data.charAt(3) != ' ') && (data.endsWith(CRLF))) {
        		//System.out.println("FTPC : readed new string : "+data) ;
        		data="";
        		repl=-1;
        	}
        } while (!data.endsWith(CRLF));
        //System.out.println("FTPC : readed final "+data);
        mdlet.addLogReply(repl);
        return repl;
	}

	private void writeString(DataOutputStream out, String v) throws IOException {
		mdlet.addLogCommand(v.substring(0,4));
		System.out.println("Write:"+v);
		out.write((v+CRLF).getBytes());
		byteout+=(v+CRLF).length();
	}
}